/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
  * © 2016 OX Software GmbH, Germany. info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/spreadsheet/utils/intervalarray', [
    'io.ox/office/tk/utils/iteratorutils',
    'io.ox/office/tk/utils/arraytemplate',
    'io.ox/office/spreadsheet/utils/interval'
], function (IteratorUtils, ArrayTemplate, Interval) {

    'use strict';

    // class IntervalArray ====================================================

    /**
     * Represents an array of column/row index intervals. The array elements
     * are instances of the class Interval.
     *
     * @constructor
     *
     * @extends Array
     */
    var IntervalArray = ArrayTemplate.create(Interval);

    // static methods ---------------------------------------------------------

    /**
     * Creates an index interval array from the passed array of single column
     * or row indexes. The index intervals in the resulting array will be
     * merged, i.e. no index interval overlaps with or adjoins to another index
     * interval. The index intervals will be sorted in ascending order.
     *
     * @param {Array<Number>} indexes
     *  An unordered array of column/row indexes to be merged to an array of
     *  index intervals.
     *
     * @returns {IntervalArray}
     *  A sorted array of unified index intervals exactly covering the passed
     *  column/row indexes.
     */
    IntervalArray.mergeIndexes = function (indexes) {

        // simplifications for small arrays
        switch (indexes.length) {
            case 0:
                return new IntervalArray();
            case 1:
                return new IntervalArray(new Interval(indexes[0]));
        }

        // the resulting intervals
        var result = new IntervalArray();

        // create the resulting intervals (expand last interval, or create a new interval)
        var interval = null;
        _.sortBy(indexes).forEach(function (index) {
            if ((interval = _.last(result)) && (index <= interval.last + 1)) {
                interval.last = index;
            } else {
                result.push(new Interval(index));
            }
        });

        return result;
    };

    // public methods ---------------------------------------------------------

    /**
     * Returns a unique string key for this array of index intervals that can
     * be used for example as key in an associative map. This method is faster
     * than the method IntervalArray.toString().
     *
     * @returns {String}
     *  A unique string key for this instance.
     */
    IntervalArray.prototype.key = function () {
        return this.map(function (interval) { return interval.key(); }).join(' ');
    };

    /**
     * Returns the number of column/row indexes covered by all index intervals
     * in this array.
     *
     * @returns {Number}
     *  The number of column/row indexes covered by all index intervals.
     */
    IntervalArray.prototype.size = function () {
        return this.reduce(function (count, interval) { return count + interval.size(); }, 0);
    };

    /**
     * Returns whether an index interval in this array contains the passed
     * column or row index.
     *
     * @param {Number} index
     *  The column or row index to be checked.
     *
     * @returns {Boolean}
     *  Whether an index interval in this array contains the passed index.
     */
    IntervalArray.prototype.containsIndex = function (index) {
        return this.some(function (interval) { return interval.containsIndex(index); });
    };

    /**
     * Returns whether all index intervals in the passed array are contained
     * completely in any of the index intervals in this array.
     *
     * @param {IntervalArray|Interval} intervals
     *  The other index intervals to be checked. This method also accepts a
     *  single index interval as parameter.
     *
     * @returns {Boolean}
     *  Whether all index intervals in the passed array are contained
     *  completely in any of the index intervals in this array.
     */
    IntervalArray.prototype.contains = function (intervals) {
        return IntervalArray.every(intervals, function (interval2) {
            return this.some(function (interval1) {
                return interval1.contains(interval2);
            });
        }, this);
    };

    /**
     * Returns whether any index interval in this array overlaps with any index
     * interval in the passed array.
     *
     * @param {IntervalArray|Interval} intervals
     *  The other index intervals to be checked. This method also accepts a
     *  single index interval as parameter.
     *
     * @returns {Boolean}
     *  Whether any index interval in this array overlaps with any index
     *  interval in the passed array.
     */
    IntervalArray.prototype.overlaps = function (intervals) {
        return this.some(function (interval1) {
            return IntervalArray.some(intervals, function (interval2) {
                return interval1.overlaps(interval2);
            });
        });
    };

    /**
     * Returns whether any pair of index intervals in this array overlap each
     * other.
     *
     * @returns {Boolean}
     *  Whether any pair of index intervals in this array overlap each other.
     */
    IntervalArray.prototype.overlapsSelf = function () {

        // simplifications for small arrays
        switch (this.length) {
            case 0:
            case 1:
                return false;
            case 2:
                return this[0].overlaps(this[1]);
        }

        // work on a sorted clone for performance (check adjacent array elements)
        return this.clone().sort().some(function (interval, index, array) {
            return (index > 0) && (array[index - 1].overlaps(interval));
        });
    };

    /**
     * Returns the bounding interval of the index intervals in this array (the
     * smallest index interval that contains all intervals in the array).
     *
     * @returns {Interval|Null}
     *  The bounding interval containing all index intervals; or null, if this
     *  interval array is empty.
     */
    IntervalArray.prototype.boundary = function () {

        switch (this.length) {
            case 0:
                return null;
            case 1:
                // this method must not return original array elements
                return this[0].clone();
        }

        var result = this[0].clone();
        this.forEach(function (interval) {
            result.first = Math.min(result.first, interval.first);
            result.last = Math.max(result.last, interval.last);
        });
        return result;
    };

    /**
     * Returns a shallow copy of this index interval array that does not
     * contain any duplicate index intervals.
     *
     * @returns {IntervalArray}
     *  A shallow copy of this index interval array without any duplicates.
     */
    IntervalArray.prototype.unify = function () {
        var result = new IntervalArray(), map = {};
        this.forEach(function (interval) {
            var key = interval.key();
            if (!map[key]) {
                result.push(interval);
                map[key] = true;
            }
        });
        return result;
    };

    /**
     * Merges the index intervals in this array so that no index interval in
     * the result overlaps with or adjoins to another index interval. The
     * resulting index intervals will be sorted in ascending order.
     *
     * @returns {IntervalArray}
     *  A sorted array of merged index intervals exactly covering the index
     *  intervals in this array.
     */
    IntervalArray.prototype.merge = function () {

        // make deep copy, and sort the intervals by start index
        var result = this.clone(true).sort();

        // merge overlapping intervals in-place
        var interval2 = null;
        result.forEach(function (interval1, index) {
            // extend current interval, as long as adjacent or overlapping with next interval
            while ((interval2 = result[index + 1]) && (interval1.last + 1 >= interval2.first)) {
                interval1.last = Math.max(interval1.last, interval2.last);
                result.splice(index + 1, 1);
            }
        });

        return result;
    };

    /**
     * Works similar to merging the index interval array (see method
     * IntervalArray.merge()) but divides the intervals that overlap each other
     * into smaller adjoining parts, so that the resulting intervals do not
     * cover the boundaries of any of the original intervals.
     *
     * Example: The interval array [A:D,C:F] will be partitioned to the
     * intervals A:B (leading part of first interval), C:D (common parts of
     * both intervals), and E:F (trailing part of second interval).
     *
     * @returns {IntervalArray}
     *  A sorted array of index intervals that do not overlap, that exactly
     *  cover the intervals in this array, and that are partitioned in a way
     *  that the resulting intervals do not cover any boundary of an original
     *  interval. Each interval object in the resulting array will contain an
     *  additional property 'coveredBy' that is an IntervalArray containing the
     *  original intervals of this instance covering that particular result
     *  interval.
     */
    IntervalArray.prototype.partition = function () {

        // sort the intervals in this array by start index
        var sourceIntervals = this.clone().sort();
        // the source intervals to be processed for the next partial interval (updated during iteration)
        var coveringIntervals = new IntervalArray();
        // the resulting partition
        var resultIntervals = new IntervalArray();

        // creates a new partial interval and inserts it into the result array
        // (this function must be defined outside thw while-loop)
        function createPartialInterval() {

            // the start index of the new partial interval to be created
            var first = sourceIntervals.empty() ? Number.POSITIVE_INFINITY : sourceIntervals.first().first;
            if (!coveringIntervals.empty()) { first = resultIntervals.last().last + 1; }

            // remove all original intervals already processed in the preceding iteration that cannot be
            // part of the partition anymore, and find the minimum end index of all remaining source intervals
            var last = Number.POSITIVE_INFINITY;
            coveringIntervals.forEachReverse(function (leadingInterval, index) {
                if (first <= leadingInterval.last) {
                    last = Math.min(last, leadingInterval.last);
                } else {
                    coveringIntervals.splice(index, 1);
                }
            });

            // add all source intervals with the current start position to the intermediate array
            while (!sourceIntervals.empty() && (first === sourceIntervals.first().first)) {
                var sourceInterval = sourceIntervals.shift();
                last = Math.min(last, sourceInterval.last);
                coveringIntervals.push(sourceInterval);
            }

            // restrict the new partial interval to the start of the next source interval not processed here
            if (!sourceIntervals.empty()) {
                last = Math.min(last, sourceIntervals.first().first - 1);
            }

            // add the resulting partial interval to the result
            if (!coveringIntervals.empty()) {
                var partInterval = new Interval(first, last);
                resultIntervals.push(partInterval);
                partInterval.coveredBy = coveringIntervals.clone();
            }
        }

        // process the source intervals
        while (!sourceIntervals.empty() || !coveringIntervals.empty()) {
            createPartialInterval();
        }

        return resultIntervals;
    };

    /**
     * Returns the index intervals covered by this array, and the passed array
     * of index intervals. More precisely, returns the existing intersection
     * intervals from all interval pairs of the cross product of the interval
     * arrays.
     *
     * @param {IntervalArray|Interval} intervals
     *  The other index intervals to be intersected with this index intervals.
     *  This method also accepts a single index interval as parameter.
     *
     * @returns {IntervalArray}
     *  The index intervals covered by this array, and the passed array of
     *  index intervals.
     */
    IntervalArray.prototype.intersect = function (intervals) {
        var result = new IntervalArray();
        this.forEach(function (interval1) {
            IntervalArray.forEach(intervals, function (interval2) {
                var interval = interval1.intersect(interval2);
                if (interval) { result.push(interval); }
            });
        });
        return result;
    };

    /**
     * Returns the difference of the index intervals in this array, and the
     * passed index intervals (all index intervals contained in this array,
     * that are not contained in the passed array).
     *
     * @param {IntervalArray|Interval} intervals
     *  The index intervals to be removed from this array. This method also
     *  accepts a single index interval as parameter.
     *
     * @returns {IntervalArray}
     *  All index intervals that are contained in this array, but not in the
     *  passed array. May be an empty array, if the passed index intervals
     *  completely cover all index intervals in this array.
     */
    IntervalArray.prototype.difference = function (intervals) {

        // the resulting intervals
        var result = this;

        // reduce intervals in 'result' with each interval from the passed array
        IntervalArray.some(intervals, function (interval2) {

            // initialize the arrays for this iteration
            var source = result;
            result = new IntervalArray();

            // process each interval in the source interval list
            source.forEach(function (interval1) {

                // check if the intervals cover each other at all
                if (!interval1.overlaps(interval2)) {
                    result.push(interval1.clone());
                    return;
                }

                // do nothing if interval2 covers interval1 completely (delete interval1 from the result)
                if (interval2.contains(interval1)) {
                    return;
                }

                // if interval1 starts before interval2, extract the leading part of interval1
                if (interval1.first < interval2.first) {
                    result.push(new Interval(interval1.first, interval2.first - 1));
                }

                // if interval1 ends after interval2, extract the trailing part of interval1
                if (interval1.last > interval2.last) {
                    result.push(new Interval(interval2.last + 1, interval1.last));
                }
            });

            // early exit the loop, if all source intervals have been deleted already
            return result.length === 0;
        });

        // ensure to return a clone
        return (result === this) ? this.clone(true) : result;
    };

    /**
     * Modifies the indexes of all index intervals in this array relatively.
     *
     * @param {Number} distance
     *  The distance to move the indexes of all index intervals in this array.
     *  Negative values will decrease the indexes.
     *
     * @returns {IntervalArray}
     *  A reference to this instance.
     */
    IntervalArray.prototype.move = function (distance) {
        this.forEach(function (interval) { interval.move(distance); });
        return this;
    };

    /**
     * Creates an iterator that visits the single indexes of the intervals in
     * this array.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Boolean} [options.reverse=false]
     *      If set to true, the intervals in this array, and the indexes in all
     *      intervals, will be visited in reversed order.
     *  @param {Number} [options.begin]
     *      If specified, only the intervals with an index greater than or
     *      equal to this value (in reversed mode: with an index less than or
     *      equal to this value) will be visited. If omitted, all intervals
     *      from the beginning (in reverse mode: from the end) of this array
     *      will be visited.
     *  @param {Number} [options.end]
     *      If specified, only the intervals with an index less than this value
     *      (in reverse mode: with an index greater than this value) will be
     *      visited (half-open interval!). If omitted, all intervals to the end
     *      (in reverse mode: to the beginning) of this array will be visited.
     *
     * @returns {Object}
     *  An iterator object that implements the standard EcmaScript iterator
     *  protocol, i.e. it provides the method next() that returns a result
     *  object with the following properties:
     *  - {Number} value
     *      The resulting index inside the interval currently visited.
     *  - {Interval} interval
     *      The original index interval from this interval array containing the
     *      index currently visited.
     *  - {Number} index
     *      The array index (!) of the interval in the property 'interval' of
     *      this result object.
     */
    IntervalArray.prototype.indexIterator = function (options) {

        // creates an interval iterator from the result of the outer array iterator
        function createIntervalIterator(interval) {
            return interval.iterator(options);
        }

        // combines the results of the outer and inner iterators
        function createIteratorResult(arrayResult, indexResult) {
            indexResult.interval = arrayResult.value;
            indexResult.index = arrayResult.index;
            return indexResult;
        }

        // the nested iterator, combining an outer array iterator with an inner interval index iterator
        return IteratorUtils.createNestedIterator(this.iterator(options), createIntervalIterator, createIteratorResult);
    };

    /**
     * Returns the string representation of the index intervals as columns.
     *
     * @param {String} [separator=',']
     *  The separator text inserted between the intervals.
     *
     * @returns {String}
     *  The upper-case string representation of the index intervals as columns.
     */
    IntervalArray.prototype.stringifyAsCols = function (separator) {
        return _.invoke(this, 'stringifyAsCols').join(separator);
    };

    /**
     * Returns the string representation of the index intervals as rows.
     *
     * @param {String} [separator=',']
     *  The separator text inserted between the intervals.
     *
     * @returns {String}
     *  The one-based string representation of the index intervals as rows.
     */
    IntervalArray.prototype.stringifyAsRows = function (separator) {
        return _.invoke(this, 'stringifyAsRows').join(separator);
    };

    /**
     * Returns the string representation of the index intervals as columns, or
     * as rows.
     *
     * @param {Boolean} columns
     *  Whether to return the column representation (true), or the row
     *  representation (false) of the index intervals.
     *
     * @param {String} [separator=',']
     *  The separator text inserted between the intervals.
     *
     * @returns {String}
     *  The string representation of this index interval.
     */
    IntervalArray.prototype.stringifyAs = function (columns, separator) {
        return columns ? this.stringifyAsCols(separator) : this.stringifyAsRows(separator);
    };

    // exports ================================================================

    return IntervalArray;

});
